package cyclops.function;

import java.util.Optional;
import java.util.concurrent.Executor;
import java.util.function.Function;

import cyclops.control.Eval;
import cyclops.async.Future;
import cyclops.control.Maybe;
import cyclops.control.Try;

public interface Fn6<T1, T2, T3, T4, T5, T6, R> {

    public R apply(T1 a, T2 b, T3 c, T4 d, T5 e, T6 f);

    default Fn1<T2, Fn1<T3, Fn1<T4, Fn1<T5, Fn1<T6, R>>>>> apply(final T1 s) {
        return Curry.curry6(this)
                    .apply(s);
    }

    default Fn1<T3, Fn1<T4, Fn1<T5, Fn1<T6, R>>>> apply(final T1 s, final T2 s2) {
        return apply(s).apply(s2);
    }

    default Fn1<T4, Fn1<T5, Fn1<T6, R>>> apply(final T1 s, final T2 s2, final T3 s3) {
        return apply(s).apply(s2)
                       .apply(s3);
    }

    default Fn1<T5, Fn1<T6, R>> apply(final T1 s, final T2 s2, final T3 s3, final T4 s4) {
        return apply(s).apply(s2)
                       .apply(s3)
                       .apply(s4);
    }

    default Fn1<T6, R> apply(final T1 s, final T2 s2, final T3 s3, final T4 s4, final T5 s5) {
        return apply(s).apply(s2)
                       .apply(s3)
                       .apply(s4)
                       .apply(s5);
    }
    default Fn6<T1, T2, T3, T4, T5, T6, Maybe<R>> lift() {
        return (s1, s2, s3, s4, s5,s6) ->  Maybe.fromLazy(Eval.later(()->Maybe.ofNullable(apply(s1,s2,s3,s4,s5,s6))));
    }
    default Fn6<T1, T2, T3, T4, T5, T6, Future<R>> lift(Executor ex) {

        return (s1, s2, s3, s4, s5,s6) -> Future.ofSupplier(() -> apply(s1, s2, s3, s4, s5,s6), ex);
    }

    default Fn6<T1, T2, T3, T4, T5, T6, Try<R, Throwable>> liftTry() {
        return (s1, s2, s3, s4, s5,s6) -> Try.withCatch(() -> apply(s1, s2, s3, s4, s5,s6), Throwable.class);
    }

    default Fn6<T1, T2, T3, T4, T5, T6, Optional<R>> liftOpt() {

        return (s1, s2, s3, s4, s5, s6) -> Optional.ofNullable(apply(s1, s2, s3, s4, s5, s6));
    }

    default Fn1<? super T1, Fn1<? super T2, Fn1<? super T3, Fn1<? super T4, Fn1<? super T5,Fn1<? super T6, ? extends R>>>>>> curry() {
        return CurryVariance.curry6(this);
    }

    default <V> Fn6<T1, T2, T3, T4, T5, T6,V> andThen(Function<? super R, ? extends V> after) {
        return (t1,t2,t3,t4,t5,t6)-> after.apply(apply(t1,t2,t3,t4,t5,t6));
    }

    public static <T1, T2, T3,T4,T5,T6,R> Fn6<T1,T2,T3,T4,T5,T6,R> λ(final Fn6<T1,T2,T3,T4,T5,T6,R> triFunc){
        return triFunc;
    }

    public static <T1, T2, T3,T4,T5,T6,R> Fn6<? super T1,? super T2,? super T3,? super T4,? super T5,? super T6,? extends R> λv(final Fn6<? super T1,? super T2,? super T3,? super T4,? super T5,? super T6,? extends R> triFunc){
        return triFunc;
    }
}
